/*
  YUY2 smart smooth plugin for Avisynth -- this plugin has been ported
  from Donald Graft's VirtualDub Smart Smooth filter to process YUY2
  frames. It removes noise from the video frame. To avoid edge blurring
  it only processes pixels that are in the color range below the given
  threshold. It is most useful for processing anime that has been recorded
  from an analog device.
  The filter algorithm has been developed by Donald Graft. Avery Lee did
  the MMX optimizations for the RGB32 VirtualDub plugin.
  The code below was developed using both, Avery Lee's MMX code and Donald
  Graft's algorithm but unlike the RGB32 version this plugin does not support
  interlaced video frames.

	Copyright (C) 2002/2003 Hans-Martin Kreuz, All Rights Reserved

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Hans-Martin Kreuz
	hmk@klose.de.
*/

#include "avisynth.h"

class SmartSmoothYuy: public GenericVideoFilter
{
  struct filterData
  {
    unsigned __int32 sumCount;
    unsigned __int16 sumY1;
    unsigned __int16 sumU;
    unsigned __int16 sumY2;
    unsigned __int16 sumV;
  };

  __int32 threshold, radius;

public:
  SmartSmoothYuy::SmartSmoothYuy(PClip _child, __int32 _threshold,
    __int32 _radius, IScriptEnvironment* env) :
    GenericVideoFilter(_child), threshold(_threshold), radius(_radius)
  {
    VideoInfo vi;

    vi = GetVideoInfo();

    if (!vi.IsYUY2())
      env->ThrowError("SmartSmoothYuy: video source must be YUY2");
    if (threshold < 1)
      env->ThrowError("SmartSmoothYuy: threshold out of range (1 - (2^31 - 1))");
    if (radius < 1)
      env->ThrowError("SmartSmoothYuy: radius out of range (radius > 0)");
  }

	PVideoFrame __stdcall GetFrame(int n,IScriptEnvironment* env);
};

PVideoFrame __stdcall SmartSmoothYuy::GetFrame(int n, IScriptEnvironment* env)
{
  PVideoFrame srcFrame, dstFrame;
	srcFrame = child->GetFrame(n, env);
  dstFrame = env->NewVideoFrame(vi);

	int xFrame, yFrame;
  // bytewise pointer to source fram (YUY2 16 bit bitmap)
	const unsigned char* srcp = srcFrame->GetReadPtr();
  // height is the number of scanlines of the frame
	const int srcFrameHeight = srcFrame->GetHeight();
  // the count of bytes that a single scanline uses
  const int srcFramePitch = srcFrame->GetPitch();
  // bytewise pointer to destination frame (YUY2 16 bit bitmap)
	unsigned char* dstp = dstFrame->GetWritePtr();
	const int dstFrameHeight = dstFrame->GetHeight();
  const int dstFramePitch = dstFrame->GetPitch();

  unsigned __int32* src = (unsigned __int32*)srcp;
  unsigned __int32* dst = (unsigned __int32*)dstp;

  // if width is 720 pixels / two pixels at once makes width = 360
  int srcFrameWidth = srcFramePitch / 4;
  int dstFrameWidth = dstFramePitch / 4;

  unsigned __int32 nucleus;
  unsigned __int32* startPixel;
  __int32 pixelCount;
  int xlo, xhi, ylo, yhi;
  int lineCount;
  filterData asmResult;
  void* resultAdr = &asmResult;
  __int32 lThr = threshold;

  // yFrame is the line counter
  for (yFrame = 0; yFrame < dstFrameHeight; yFrame++)
	{
    // xFrame is the two pixel counter (4 bytes = 2 pixels)
    // dstFrameRowSize is the picture width in bytes
    // on a frame with a width of 720 YUY2 pixels xFrame will count from 0 to
    // 359
		for (xFrame = 0; xFrame < dstFrameWidth; xFrame++)
    {
      nucleus = src[xFrame];

      asmResult.sumCount = 0;
      asmResult.sumY1 = 0;
      asmResult.sumU = 0;
      asmResult.sumY2 = 0;
      asmResult.sumV = 0;

      // check if we are outside of the frame's borders and if so, correct the
      // boundaries
      if ((xlo = xFrame - radius) < 0) xlo = 0;
      if ((xhi = xFrame + radius) >= dstFrameWidth) xhi = dstFrameWidth - 1;
      pixelCount = xlo - xhi - 1;
      // since we process two pixels at once in x direction we use the double
      // radius towards the y direction to process a square and not a rectangle
      if ((ylo = yFrame - radius * 2) < 0) ylo = 0;
      if ((yhi = yFrame + radius * 2) >= dstFrameHeight)
        yhi = dstFrameHeight - 1;
      lineCount = yhi - ylo + 1;
      startPixel = ((unsigned __int32*)srcp) + xlo + ylo * srcFrameWidth;

      // outer block loop (counts lines of a block)
      do
      {
        // calculate average for one line of the square
        __asm
        {
          // mm0 <- nucleus
          // mm1 <- compareVal (upper 32 bit)
          // mm2 <- pixels to be processed
          // mm3 <- count of pixel values added together
          // mm4 <- sum of added pixel values
          pxor  mm7,mm7
          movd  mm0,nucleus
          punpcklbw mm0,mm7
          movd  mm1,lThr
          psllq mm1,32
          mov ebx,resultAdr
          movd  mm3,[ebx]
          movq  mm4,[ebx + 4]
          mov ecx,pixelCount
          mov edx,startPixel

          // start processing
        nextPixel:
          movd  mm2,[edx]
          punpcklbw mm2,mm7
          movq  mm6,mm2
          psubw mm6,mm0
          pmaddwd mm6,mm6
          punpckldq mm5,mm6
          paddd mm6,mm5
          movq  mm5,mm1
          pcmpgtd mm5,mm6
          punpckhdq mm5,mm5
          pand  mm2,mm5
          paddw mm4,mm2
          psrlq mm5,63
          paddd mm3,mm5

          add edx,4
          inc ecx
          jne nextPixel

          movd  [ebx],mm3
          movq  [ebx + 4],mm4
        };
        
        startPixel += srcFrameWidth;
      }
      while (--lineCount);

      __asm emms;

      // set the new pixel to the average value of the pixel sum
      asmResult.sumY1 = (asmResult.sumY1 + (__int16)asmResult.sumCount / 2)
        / (__int16)asmResult.sumCount;
      asmResult.sumU = (asmResult.sumU + (__int16)asmResult.sumCount / 2)
        / (__int16)asmResult.sumCount;
      asmResult.sumY2 = (asmResult.sumY2 + (__int16)asmResult.sumCount / 2)
        / (__int16)asmResult.sumCount;
      asmResult.sumV = (asmResult.sumV + (__int16)asmResult.sumCount / 2)
        / (__int16)asmResult.sumCount;

      dst[xFrame] = (asmResult.sumV << 24) + (asmResult.sumY2 << 16)
        + (asmResult.sumU << 8) + asmResult.sumY1;
    }

    // next row
    src += srcFrameWidth;
    dst += dstFrameWidth;
	}

  return dstFrame;
}

AVSValue __cdecl Create_SmartSmoothYuy(AVSValue args, void* user_data,
  IScriptEnvironment* env)
{
  __int32 threshold = 225;
  __int32 radius = 3;

  return new SmartSmoothYuy(args[0].AsClip(),
    args[1].AsInt(threshold), // pixel add to average sum threshold
    args[2].AsInt(radius), // search square radius
    env);
}

extern "C" __declspec(dllexport)
  const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env)
{
  env->AddFunction("SmartSmoothYuy", "c[threshold]i[radius]i",
    Create_SmartSmoothYuy, 0);
	return "Smart smooth plugin for YUY2 frames";
}
